/***
 * Copyright (C) 1997,98 -- Miroslav Zagorac  <zaga@fly.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "lib_debug.h"
#include "lib_image.h"
#include "lib_options.h"
#include "lib_pcx.h"
#include "lib_string.h"
#include "lib_vga.h"

#ifdef USE_JPEGLIB
#  include "lib_jpeg.h"
#endif


/***
 * Ako dvije slike koriste istu datoteku, onda je glupo svaki put datoteku
 * otvarati, alocirati memoriju, i gubiti vrijeme/memoriju za slicne stvari.
 * Zato ovdje imamo polje pointera u koje se upisuju imena datoteka, kao i
 * adrese alocirane memorije za svaku datoteku posebno. Time se postize vrlo
 * efikasna usteda memorije. Jedino na sto treba paziti je da svaka datoteka
 * ima apsolutno istu paletu.
 */
static struct IMAGES_DATA
{
   char              *filename;
   struct IMAGE_DATA *image_ptr;
} images[MAX_IMAGES];
static short images_number = 0;


#ifdef DEBUG

#ifdef ANSI_FUNC

static void debug_image (struct IMAGE_DATA *image)
#else

static void debug_image (image)
   struct IMAGE_DATA *image;
#endif
{
   debug (("image <0x%08lx> = {\n", (long) image))
   debug (("   width   = %d\n", image->width))
   debug (("   height  = %d\n", image->height))
   debug (("   image   = 0x%08lx\n", (long) image->image))
   debug (("   palette = 0x%08lx\n", (long) image->palette))
   debug ((" }\n"))
}

#endif /* DEBUG */


#ifdef ANSI_FUNC

static short image_format (char *filename)
#else
   
static short image_format (filename)
   char *filename;
#endif
{
   char buffer[BUFSIZ];

   debug (("{%s(\"%s\")", __FUNCTION__, filename))

   if (strrchr (filename, '.'))
   {
      (void) strcpy (buffer, strrchr (filename, '.') + 1);
      (void) strlwr (buffer);
   }
   else
      (void) strcpy (buffer, filename);

#ifdef USE_JPEGLIB
   if (!strcmp (buffer, "jpeg") || !strcmp (buffer, "jpg"))
      debug_return ("} == %d", IF_JPEG)
   else
#endif
   if (! strcmp (buffer, "pcx"))
      debug_return ("} == %d", IF_PCX)
   else if (! strcmp (buffer, "vga"))
      debug_return ("} == %d", IF_VGA)
   else if (! strcmp (buffer, "raw"))
      debug_return ("} == %d", IF_RAW)

   debug (("Unrecognized image extension: \"%s\"\n", buffer))

   debug_return ("} == %d", FUNC_RET_ERROR)
}


#ifdef ANSI_FUNC

void image_error (char *mesg, int errnum, struct IMAGE_DATA **image, FILE *vga_fd)
#else

void image_error (mesg, errnum, image, vga_fd)
   char *mesg;
   int errnum;
   struct IMAGE_DATA **image;
   FILE *vga_fd;
#endif
{
   debug (("{%s(\"%s\", %d, 0x%08lx, 0x%08lx)", __FUNCTION__, mesg, errnum, (long) image, (long) vga_fd))
   debug (("%s%s%s\n", mesg, errnum ? ": " : "", errnum ? strerror (errnum) : ""))

   if (verbose_level)
      (void) fprintf (stderr, "%s%s%s\n", mesg, errnum ? ": " : "", errnum ? strerror (errnum) : "");

   if (vga_fd)
      (void) fclose (vga_fd);

   image_free (image);

   debug_endfunc
}


#ifdef ANSI_FUNC

struct IMAGE_DATA *image_alloc (short width, short height)
#else

struct IMAGE_DATA *image_alloc (width, height)
   short width;
   short height;
#endif
{
   struct IMAGE_DATA *image;

   debug (("{%s(%d, %d)", __FUNCTION__, width, height))

   if (! (image = (struct IMAGE_DATA *) calloc (1, sizeof (struct IMAGE_DATA))))
   {
      image_error ("calloc(IMAGE_DATA)", errno, &image, NULL);

      debug_return ("} == 0x%08x", NULL)
   }

   if (! (image->image = (image_t *) calloc (width * height, sizeof (image_t))))
   {
      image_error ("calloc(image)", errno, &image, NULL);

      debug_return ("} == 0x%08x", NULL)
   }

   if (! (image->palette = (palette_t *) calloc (SIZEOF_PALETTE, sizeof (palette_t))))
   {
      image_error ("calloc(palette)", errno, &image, NULL);

      debug_return ("} == 0x%08x", NULL)
   }

   image->width  = width;
   image->height = height;

#ifdef DEBUG
   debug_image (image);
#endif

   debug_return ("} == 0x%08x", image)
}


#ifdef ANSI_FUNC

void image_free (struct IMAGE_DATA **image)
#else

void image_free (image)
   struct IMAGE_DATA **image;
#endif
{
   debug (("{%s(%08lx)", __FUNCTION__, (long) image))

   if (image && *image)
   {
      short i;

      debug (("Freeing memory at: image=0x%08lx, palette=0x%08lx\n", (long) (*image)->image, (long) (*image)->palette))

      for (i = 0; images_number && (i < MAX_IMAGES); i++)
         if (images[i].image_ptr == *image)
         {
            images[i].filename  = NULL;
            images[i].image_ptr = NULL;
            images_number--;

            debug (("deleted entry images[%d]\n", i))
         }

      if ((*image)->palette)
         free ((*image)->palette);

      if ((*image)->image)
         free ((*image)->image);

      free (*image);

      *image = NULL;
   }
   else
      debug (("Memory already deleted\n"))

   debug_endfunc
}


#ifdef ANSI_FUNC

struct IMAGE_DATA *image_get_part (struct IMAGE_DATA *image_src, short x, short y, short width, short height)
#else

struct IMAGE_DATA *image_get_part (image_src, x, y, width, height)
   struct IMAGE_DATA *image_src;
   short x;
   short y;
   short width;
   short height;
#endif
{
   struct IMAGE_DATA *image;
   short              i;

   debug (("{%s(0x%08lx, %d, %d, %d, %d)", __FUNCTION__, (long) image_src, x, y, width, height))

#ifdef DEBUG
   debug_image (image_src);
#endif

   if (image_src == NULL)
      debug_return ("} == 0x%08x", NULL)

   if ((image = image_alloc (width, height)) == NULL)
      debug_return ("} == 0x%08x", NULL)

   if (image_src->palette)
      (void) memcpy (image->palette, image_src->palette, SIZEOF_PALETTE);

   for (i = 0; i < height; i++)
      (void) memcpy (image->image + i * image->width, image_src->image + x + (y + i) * image_src->width, image->width);

   debug_return ("} == 0x%08x", image)
}


struct IMAGE_DATA *image_crop (struct IMAGE_DATA *image)
{
   struct IMAGE_DATA *retval = NULL;

   debug (("{%s(0x%08lx)", __FUNCTION__, (long) image))

   if (image && image->image)
   {
      short   x_0 = -1L, x_1 = -1, y_0 = -1L, y_1 = -1L, x, y;
      image_t crop_value = *image->image;

      /***
       * crop_y
       */
      for (y = 0L; y < (image->height / 2); y++)
      {
         if (y_0 < 0L)
            for (x = 0L; x < image->width; x++)
               if (image->image[x + y * image->width] != crop_value)
               {
                  y_0 = y;

                  break;
               }

         if (y_1 < 0L)
            for (x = 0L; x < image->width; x++)
               if (image->image[x + (image->height - y - 1) * image->width] != crop_value)
               {
                  y_1 = image->height - y - 1;

                  break;
               }
      }

      y_0 = (y_0 < 0) ? 0 : y_0;
      y_1 = (y_1 < 0) ? image->height : y_1;

      /***
       * crop_x
       */
      for (x = 0L; x < (image->width / 2); x++)
      {
         if (x_0 < 0L)
            for (y = y_0; y < y_1; y++)
               if (image->image[x + y * image->width] != crop_value)
               {
                  x_0 = x;

                  break;
               }

         if (x_1 < 0L)
            for (y = y_0; y < y_1; y++)
               if (image->image[image->width - x - 1 + y * image->width] != crop_value)
               {
                  x_1 = image->width - x - 1;

                  break;
               }
      }

      x_0 = (x_0 < 0) ? 0 : x_0;
      x_1 = (x_1 < 0) ? image->width : x_1;

      if ((x_1 != 0) || (x_1 != image->width) || (y_1 != 0) || (y_1 != image->height))
         retval = image_get_part (image, x_0, y_0, x_1 - x_0 + 1, y_1 - y_0 + 1);
   }

   debug_return ("} == 0x%08x", retval)
}


#ifdef ANSI_FUNC

struct IMAGE_DATA *image_rotate (struct IMAGE_DATA *image, double scale, double angle, short crop_flag)
#else

struct IMAGE_DATA *image_rotate (image, scale, angle, crop_flag)
   struct IMAGE_DATA *image:
   double scale:
   double angle:
   short crop_flag;
#endif
{
   struct IMAGE_DATA *new_image, *new_image_crop;
   image_t           *new_image_ptr;
   short              new_width = sqrt (image->height * image->height + image->width * image->width) * scale + 1L, new_height = new_width;
   short              x, y;
   long               sinas = sin (-angle) * 65536L / scale;
   long               cosas = cos (-angle) * 65536L / scale;
   /***
    * x' = cos(-angle) + sin(-angle)
    * y' = cos(-angle) - sin(-angle)
    */
   long               xc = image->width * 32768L - new_height / 2 * (cosas + sinas);
   long               yc = image->height * 32768L - new_width / 2 * (cosas - sinas);

   debug (("{%s(0x%08lx, %4.2f, %4.2f)", __FUNCTION__, (long) image, scale, angle))

   if ((new_image = image_alloc (new_width, new_height)) == NULL)
      debug_return ("} == 0x%08x", NULL)

   if (image->palette)
      (void) memcpy (new_image->palette, image->palette, SIZEOF_PALETTE);

   new_image_ptr = new_image->image;

   for (y = 0; y < new_height; y++)
   {
      long tx = xc, ty = yc;

      for (x = 0; x < new_width; x++)
      {
         long x_tmp = tx >> 16, y_tmp = ty >> 16;

         if ((x_tmp >= 0) && (x_tmp < image->width) && (y_tmp >= 0) && (y_tmp < image->height))
            new_image_ptr[x] = image->image[x_tmp + y_tmp * image->width];

         tx += cosas;
         ty -= sinas;
      }

      new_image_ptr += new_image->width;
      xc            += sinas;
      yc            += cosas;
   }

   if (crop_flag)
   {
      new_image_crop = image_crop (new_image);

      image_free (&new_image);

      debug_return ("} == 0x%08x", new_image_crop)
   }

   debug_return ("} == 0x%08x", new_image)
}


#ifdef ANSI_FUNC

struct IMAGE_DATA *image_load_raw (char *filename, palette_t *palette, off_t width, off_t height, off_t image_offset, off_t palette_offset)
#else
   
struct IMAGE_DATA *image_load_raw (filename, palette, width, height, image_offset, palette_offset)
   char *filename;
   palette_t *palette;
   off_t width;
   off_t height;
   off_t image_offset;
   off_t palette_offset;
#endif
{
   debug (("{%s(\"%s\", 0x%08lx, %ld, %ld, %ld, %ld)", __FUNCTION__, filename, (long) palette, width, height, image_offset, palette_offset))

   debug_return ("} == 0x%08x", NULL)
}


#ifdef ANSI_FUNC

struct IMAGE_DATA *image_load (char *filename, palette_t *palette, short load_flag)
#else

struct IMAGE_DATA *image_load (filename, palette, load_flag)
   char *filename;
   palette_t *palette;
   short load_flag;
#endif
/***
 * U ovoj funkciji se ucitavaju slike u PCX/JPG formatu u memoriju.
 * Ako je slika sa trazenim imenom vec bila ucitana, nece se ponovno ucitavati,
 * vec ce se samo malo petljati sa pointerima. :)
 */
{
   static struct IMAGE_DATA *image;
   short                     image_index = 0, image_fmt;

   debug (("{%s(\"%s\", 0x%08lx)", __FUNCTION__, filename, (long) palette))

   if (load_flag)
   {
      short i;

      for (i = 0; i < MAX_IMAGES; i++)
         if (images[i].image_ptr)
         {
            if (! strcmp (images[i].filename, filename))
            {
               debug (("%s: Memory already allocated at: 0x%08lx: image=0x%08lx, palette=0x%08lx\n", images[i].filename, (long) images[i].image_ptr, (long) images[i].image_ptr->image, (long) images[i].image_ptr->palette))

               debug_return ("} == 0x%08x", images[i].image_ptr)
            }
         }
         else if (image_index < 0)
            image_index = i;
   }

   image_fmt = image_format (filename);

#ifdef USE_JPEGLIB
   if (image_fmt == IF_JPEG)
      image = jpeg_image_load (filename, palette, 1.0);
   else
#endif
   if (image_fmt == IF_PCX)
      image = pcx_image_load (filename);
   else if (image_fmt == IF_VGA)
      image = vga_image_load (filename, palette);
   else
      debug_return ("} == 0x%08x", NULL)

   if (image == NULL)
      debug_return ("} == 0x%08x", NULL)

   debug (("%s: Allocated memory at: 0x%08lx\n", filename, (long) image))

   if (load_flag && (image_index >= 0))
   {
      images[image_index].filename  = strdup (filename);
      images[image_index].image_ptr = image;

      images_number++;
   }

   debug_return ("} == 0x%08x", image)
}


#ifdef ANSI_FUNC

palette_t *image_palette_load (char *filename)
#else
   
palette_t *image_palette_load (char *filename)
   char *filename;
#endif
{
   static palette_t  palette[SIZEOF_PALETTE];
   FILE             *palette_fd;

   debug (("{%s(\"%s\")", __FUNCTION__, filename))

   if ((palette_fd = fopen (filename, "r")))
   {
      char row[BUFSIZ];
      int  color_index = 0, r, g, b;

      (void) memset (palette, 0, SIZEOF_PALETTE);

      while (fgets (row, BUFSIZ, palette_fd))
         if (sscanf (row, "%d%d%d", &r, &g, &b) == 3)
         {
            palette[color_index++] = (unchar) r;
            palette[color_index++] = (unchar) g;
            palette[color_index++] = (unchar) b;
         }

      (void) fclose (palette_fd);

      debug_return ("} == 0x%08x", palette)
   }

   debug_return ("} == 0x%08x", NULL)
}


#ifdef ANSI_FUNC

short image_save (char *filename, struct IMAGE_DATA *image)
#else

short image_save (filename, image)
   char *filename;
   struct IMAGE_DATA *image;
#endif
{
   short retval = FUNC_RET_OK;

   debug (("{%s(\"%s\", 0x%08lx)", __FUNCTION__, filename, (long) image))

   if (image)
   {
      short image_fmt = image_format (filename);

#ifdef USE_JPEGLIB
      if (image_fmt == IF_JPEG)
         retval = jpeg_image_save (image, filename);
      else
#endif
      if (image_fmt == IF_PCX)
         retval = pcx_image_save (image, filename);
      else if (image_fmt == IF_VGA)
         retval = vga_image_save (image, filename);
      else
         retval = FUNC_RET_ERROR;
   }
   else
      retval = FUNC_RET_ERROR;

   debug_return ("} == %d", retval)
}


#ifdef ANSI_FUNC

void image_palette_brightness (palette_t *palette, double factor)
#else

void image_palette_brightness (palette, factor)
   palette_t *palette;
   double factor;
#endif
{
   short i;

   debug (("{%s(0x%08lx, %f)", __FUNCTION__, (long) palette, factor))

   for (i = 0; i < SIZEOF_PALETTE; i++)
   {
      short value = palette[i] * factor;

      palette[i] = (value > 255) ? 255 : value;
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

void image_clean (void)
#else

void image_clean ()
#endif
{
   short i;

   debug (("{%s()", __FUNCTION__))

   for (i = 0; images_number && (i < MAX_IMAGES); i++)
      if (images[i].image_ptr)
      {
         free (images[i].filename);
         image_free (&images[i].image_ptr);

         images[i].filename = NULL;
         images_number--;

         debug (("deleted entry images[%d]\n", i))
      }

   debug_endfunc
}
